==> sagua_0.1-1.tar.gz: sagua-0.1/src/tag_mp3.cc <==
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <unistd.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>

#include <mad.h>
#include <id3tag.h>

#include "../include/tag_mp3.h"
#include "../include/misc.h"
#include "../include/trace.h"

MP3Tag::MP3Tag(const std::string& filename)
: filename_(filename),
    validFile_(false),
    id3File_(NULL),
    id3Tag_(NULL),
    xingFlagFrames_(0x00000001L),
    xingFlags_(0),
    xingFrames_(0)
{
    Tracer::Trace trace("MP3Tag::MP3Tag");

    id3File_ = id3_file_open(filename.c_str(), ID3_FILE_MODE_READONLY);

    if(id3File_ != NULL) {
	id3Tag_ = id3_file_tag(id3File_);

	if(id3Tag_ != NULL)
	    validFile_ = true;
	else
	    trace.Debug("Can't read tag");    

    } else
	trace.Debug("Can't open file");

}

MP3Tag::~MP3Tag()
{
    if(validFile_) {
	id3_tag_delete(id3Tag_);
	id3_file_close(id3File_);
    }
}

bool MP3Tag::XingParse(struct mad_bitptr ptr, unsigned int bitlen)
{
    xingFlags_ = 0;
    
    if(bitlen < 64 || mad_bit_read(&ptr, 32) != 
	    (('X' << 24) | ('i' << 16) | ('n' << 8) | 'g') ) {

	xingFlags_ = 0;

	return false; 
    }

    xingFlags_ = mad_bit_read(&ptr, 32);
    bitlen -= 64;

    if (xingFlags_ & xingFlagFrames_) {
	if (bitlen < 32) {

	    xingFlags_ = 0;

	    return false;
	}

	xingFrames_ = mad_bit_read(&ptr, 32);

	return true;
    }

    return false;
}

Glib::ustring MP3Tag::ProcessFrame(const char* id)
{
    if(validFile_)
	if(struct id3_frame* frame = id3_tag_findframe(id3Tag_, id, 0)) {
	    union id3_field const * field = id3_frame_field(frame, 1);
	    const unsigned int nstrings = id3_field_getnstrings(field);

	    id3_ucs4_t const * ucs4;
	    id3_latin1_t* latin1;
	    Glib::ustring tmp;
	    tmp.reserve(32);

	    for(unsigned int i = 0; i < nstrings; ++i)
	    {
		ucs4 = id3_field_getstrings(field, i);

		if(!ucs4)
		    break;		

		if(strcmp(id, ID3_FRAME_GENRE) == 0)
		    ucs4 = id3_genre_name(ucs4); 

		latin1 = id3_ucs4_latin1duplicate(ucs4);    

		if(latin1 == 0)
		    break;

		tmp += reinterpret_cast<char*>(latin1);

		free(latin1);
	    }

	    if(StringIsSpace(tmp)) 
		if(strcmp(id, ID3_FRAME_TITLE) == 0)
		    return StringToUTF8(Glib::path_get_basename(filename_), "");
		else
		    return Glib::ustring();        


	    if(strcmp(id, ID3_FRAME_TITLE) == 0)
		return StringToUTF8(tmp, Glib::path_get_basename(filename_)); 
	    else
		return StringToUTF8(tmp, "");	
	} 

    if(strcmp(id, ID3_FRAME_TITLE) == 0)
	return StringToUTF8(Glib::path_get_basename(filename_), "");
    else
	return Glib::ustring();	
}

void MP3Tag::ProcessAll()
{

}

void MP3Tag::ProcessYear()
{
    year_ = ProcessFrame(ID3_FRAME_YEAR);
}

void MP3Tag::ProcessTrack()
{
    track_ = ProcessFrame(ID3_FRAME_TRACK);
}

void MP3Tag::ProcessTitle()
{
    title_ = ProcessFrame(ID3_FRAME_TITLE);
}

void MP3Tag::ProcessAlbum()
{
    album_ = ProcessFrame(ID3_FRAME_ALBUM);
}

void MP3Tag::ProcessGenre()
{
    genre_ = ProcessFrame(ID3_FRAME_GENRE);
}

void MP3Tag::ProcessArtist()
{
    artist_ = ProcessFrame(ID3_FRAME_ARTIST);
}

void MP3Tag::ProcessComment()
{
    if(validFile_) {
	unsigned int index = 0;
	struct id3_frame* frame;
	id3_ucs4_t const * ucs4;

	while(frame = id3_tag_findframe(id3Tag_, ID3_FRAME_COMMENT, index++)) {
	    ucs4 = id3_field_getstring(id3_frame_field(frame, 2));

	    if(!ucs4) {
		comment_ = "";
		return;
	    }

	    if(*ucs4)
		continue;

	    ucs4 = id3_field_getfullstring(id3_frame_field(frame, 3));	

	    if(!ucs4) {
		comment_ = "";
		return;
	    }

	    id3_latin1_t* latin1 = id3_ucs4_latin1duplicate(ucs4);	

	    Glib::ustring tmp = reinterpret_cast<char*>(latin1);

	    free(latin1);

	    comment_ = StringToUTF8(tmp, "");
	    return;
	}
    }

    comment_ = "";
}

void MP3Tag::ProcessTotalTime()
{
    Tracer::Trace trace("MP3Tag::ProcessTotalTime");

    int fd;
    struct stat stat;
    void* fdm;

    fd = open(filename_.c_str(), O_RDONLY); 
    if(fd == -1) {
	trace.Debug("Can't open file");
	
	return;
    }

    if(fstat(fd, &stat) == -1) {
	trace.Debug("Can't read file stat");
	close(fd);
	
	return;	
    }

    if(!S_ISREG(stat.st_mode)) {
	trace.Debug("File isn't regular file");
	close(fd);
	
	return;
    }

    fdm = mmap(0, stat.st_size, PROT_READ, MAP_SHARED, fd, 0);	

    if(fdm) {
	mad_timer_t duration = mad_timer_zero;
	struct mad_stream stream;
	struct mad_header header;
	struct mad_frame frame;

	mad_stream_init(&stream);	
	mad_header_init(&header);
	mad_frame_init(&frame);
	mad_stream_buffer(&stream, static_cast<unsigned char*>(fdm), stat.st_size);

	unsigned int count = 0;
	double averageBitrate = 0.0;
	FileType fileType = VBR;

	for(;;) {
	    if(mad_header_decode(&header, &stream) == -1) {
		if(MAD_RECOVERABLE(stream.error)) 
		    continue;
		else
		    break;
	    }

	    if(count == 0) {
		// You need to use mad_frame_decode to find ancillary data
		// that xing header in.
		frame.header = header;

		if(mad_frame_decode(&frame, &stream) != -1) {
		    if(XingParse(stream.anc_ptr, stream.anc_bitlen)) {

			fileType = XING_VBR;

			break;
		    }
		}

	    }

	    // may be 5 second... ( 38 * 5 = 190)
	    if(count < 190) {
		count++;
		averageBitrate += static_cast<double>(header.bitrate);
		averageBitrate /= 2;
	    } else if(count == 190) {
		// for never enter this statement 
		count++;
		// average bitrate == current bitrate, maybe constant bitrate
		if(averageBitrate == static_cast<double>(header.bitrate)) {
		    fileType = CBR;

		    break; 
		}
	    }

	    mad_timer_add(&duration, header.duration);
	}

	char timeStr_[18];

	switch(fileType) {
	    case CBR:
		trace.Debug("time calcurated in CBR");
		
		// this is not accuracy, but fast and can express almost exact time.
		totalTimeSeconds_ = static_cast<unsigned long>( static_cast<double>(stat.st_size) * 8.0 / static_cast<double>(header.bitrate) );
		snprintf(timeStr_, 18, "%02u:%02u", static_cast<unsigned int>(totalTimeSeconds_ / 60), static_cast<unsigned int>(totalTimeSeconds_ % 60));

		break;
	    
	    case VBR:
		trace.Debug("time calcurated in VBR");
		
		duration = mad_timer_abs(duration);
		totalTimeSeconds_ = duration.seconds;
		snprintf(timeStr_, 18, "%02u:%02u", static_cast<unsigned int>(duration.seconds / 60), static_cast<unsigned int>(duration.seconds % 60));
	    
		break;
		
	    case XING_VBR:
		trace.Debug("time calculated in Xing header VBR");
		
		mad_timer_multiply(&header.duration, xingFrames_);
		duration = header.duration;
		duration = mad_timer_abs(duration);
		totalTimeSeconds_ = duration.seconds;
		snprintf(timeStr_, 18, "%02u:%02u", static_cast<unsigned int>(duration.seconds / 60), static_cast<unsigned int>(duration.seconds % 60));
	    
		break;		
	}

	totalTime_ = timeStr_;

	mad_frame_finish(&frame);
	mad_header_finish(&header);
	mad_stream_finish(&stream);

	munmap(fdm, stat.st_size);
	
    } else
	trace.Debug("Can't map file");

    close(fd);

}

==> sagua_0.1.1-1.tar.gz: sagua-0.1.1/src/tag_mp3.cc <==
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <unistd.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>

#include <mad.h>
#include <id3tag.h>

#include "../include/tag_mp3.h"
#include "../include/misc.h"
#include "../include/trace.h"

MP3Tag::MP3Tag(const std::string& filename)
: filename_(filename),
    validFile_(false),
    id3File_(NULL),
    id3Tag_(NULL),
    xingFlagFrames_(0x00000001L),
    xingFlags_(0),
    xingFrames_(0)
{
    Tracer::Trace trace("MP3Tag::MP3Tag");

    id3File_ = id3_file_open(filename.c_str(), ID3_FILE_MODE_READONLY);

    if(id3File_ != NULL) {
	id3Tag_ = id3_file_tag(id3File_);

	if(id3Tag_ != NULL)
	    validFile_ = true;
	else
	    trace.Debug("Can't read tag");    

    } else
	trace.Debug("Can't open file");

}

MP3Tag::~MP3Tag()
{
    if(validFile_) {
	id3_tag_delete(id3Tag_);
	id3_file_close(id3File_);
    }
}

bool MP3Tag::XingParse(struct mad_bitptr ptr, unsigned int bitlen)
{
    xingFlags_ = 0;
    
    if(bitlen < 64 || mad_bit_read(&ptr, 32) != 
	    (('X' << 24) | ('i' << 16) | ('n' << 8) | 'g') ) {

	xingFlags_ = 0;

	return false; 
    }

    xingFlags_ = mad_bit_read(&ptr, 32);
    bitlen -= 64;

    if (xingFlags_ & xingFlagFrames_) {
	if (bitlen < 32) {

	    xingFlags_ = 0;

	    return false;
	}

	xingFrames_ = mad_bit_read(&ptr, 32);

	return true;
    }

    return false;
}

Glib::ustring MP3Tag::ProcessFrame(const char* id)
{
    if(validFile_)
	if(struct id3_frame* frame = id3_tag_findframe(id3Tag_, id, 0)) {
	    union id3_field const * field = id3_frame_field(frame, 1);
	    const unsigned int nstrings = id3_field_getnstrings(field);

	    id3_ucs4_t const * ucs4;
	    id3_latin1_t* latin1;
	    Glib::ustring tmp;
	    tmp.reserve(32);

	    for(unsigned int i = 0; i < nstrings; ++i)
	    {
		ucs4 = id3_field_getstrings(field, i);

		if(!ucs4)
		    break;		

		if(strcmp(id, ID3_FRAME_GENRE) == 0)
		    ucs4 = id3_genre_name(ucs4); 

		latin1 = id3_ucs4_latin1duplicate(ucs4);    

		if(latin1 == 0)
		    break;

		tmp += reinterpret_cast<char*>(latin1);

		free(latin1);
	    }

	    if(StringIsSpace(tmp)) 
		if(strcmp(id, ID3_FRAME_TITLE) == 0)
		    return StringToUTF8(Glib::path_get_basename(filename_), "");
		else
		    return Glib::ustring();        


	    if(strcmp(id, ID3_FRAME_TITLE) == 0)
		return StringToUTF8(tmp, Glib::path_get_basename(filename_)); 
	    else
		return StringToUTF8(tmp, "");	
	} 

    if(strcmp(id, ID3_FRAME_TITLE) == 0)
	return StringToUTF8(Glib::path_get_basename(filename_), "");
    else
	return Glib::ustring();	
}

void MP3Tag::ProcessAll()
{

}

void MP3Tag::ProcessYear()
{
    year_ = ProcessFrame(ID3_FRAME_YEAR);
}

void MP3Tag::ProcessTrack()
{
    track_ = ProcessFrame(ID3_FRAME_TRACK);
}

void MP3Tag::ProcessTitle()
{
    title_ = ProcessFrame(ID3_FRAME_TITLE);
}

void MP3Tag::ProcessAlbum()
{
    album_ = ProcessFrame(ID3_FRAME_ALBUM);
}

void MP3Tag::ProcessGenre()
{
    genre_ = ProcessFrame(ID3_FRAME_GENRE);
}

void MP3Tag::ProcessArtist()
{
    artist_ = ProcessFrame(ID3_FRAME_ARTIST);
}

void MP3Tag::ProcessComment()
{
    if(validFile_) {
	unsigned int index = 0;
	struct id3_frame* frame;
	id3_ucs4_t const * ucs4;

	while(frame = id3_tag_findframe(id3Tag_, ID3_FRAME_COMMENT, index++)) {
	    ucs4 = id3_field_getstring(id3_frame_field(frame, 2));

	    if(!ucs4) {
		comment_ = "";
		return;
	    }

	    if(*ucs4)
		continue;

	    ucs4 = id3_field_getfullstring(id3_frame_field(frame, 3));	

	    if(!ucs4) {
		comment_ = "";
		return;
	    }

	    id3_latin1_t* latin1 = id3_ucs4_latin1duplicate(ucs4);	

	    Glib::ustring tmp = reinterpret_cast<char*>(latin1);

	    free(latin1);

	    comment_ = StringToUTF8(tmp, "");
	    return;
	}
    }

    comment_ = "";
}

void MP3Tag::ProcessTotalTime()
{
    Tracer::Trace trace("MP3Tag::ProcessTotalTime");

    int fd;

    fd = open(filename_.c_str(), O_RDONLY); 
    if(fd == -1) {
	trace.Debug("Can't open file");
	
	return;
    }

    struct stat stat;

    if(fstat(fd, &stat) == -1) {
	trace.Debug("Can't read file stat");
	close(fd);
	
	return;	
    }

    if(!S_ISREG(stat.st_mode)) {
	trace.Debug("File isn't regular file");
	close(fd);
	
	return;
    }

    void* fdm;

    fdm = mmap(0, stat.st_size, PROT_READ, MAP_SHARED, fd, 0);	

    if(fdm) {
	mad_timer_t duration = mad_timer_zero;
	struct mad_stream stream;
	struct mad_header header;
	struct mad_frame frame;

	mad_stream_init(&stream);	
	mad_header_init(&header);
	mad_frame_init(&frame);
	mad_stream_buffer(&stream, static_cast<unsigned char*>(fdm), stat.st_size);

	unsigned int count = 0;
	double averageBitrate = 0.0;
	FileType fileType = VBR;

	for(;;) {
	    if(mad_header_decode(&header, &stream) == -1) {
		if(MAD_RECOVERABLE(stream.error)) 
		    continue;
		else
		    break;
	    }

	    if(count == 0) {
		// You need to use mad_frame_decode to find ancillary data
		// that xing header in.
		frame.header = header;

		if(mad_frame_decode(&frame, &stream) != -1) {
		    if(XingParse(stream.anc_ptr, stream.anc_bitlen)) {

			fileType = XING_VBR;

			break;
		    }
		}

	    }

	    // may be 5 second... ( 38 * 5 = 190)
	    if(count < 190) {
		count++;
		averageBitrate += static_cast<double>(header.bitrate);
		averageBitrate /= 2;
	    } else if(count == 190) {
		// for never enter this statement 
		count++;
		// current bitrate == average bitrate, maybe constant bitrate
		if(header.bitrate == static_cast<unsigned long>(averageBitrate)) {    
		    fileType = CBR;

		    break; 
		}
	    }

	    mad_timer_add(&duration, header.duration);
	}

	char timeStr_[18];

	switch(fileType) {
	    case CBR:
		trace.Debug("time calcurated in CBR");
		
		// this is not accuracy, but fast and can express almost exact time.
		totalTimeSeconds_ = static_cast<unsigned long>( static_cast<double>(stat.st_size) * 8.0 / static_cast<double>(header.bitrate) );
		snprintf(timeStr_, 18, "%02u:%02u", static_cast<unsigned int>(totalTimeSeconds_ / 60), static_cast<unsigned int>(totalTimeSeconds_ % 60));

		break;
	    
	    case VBR:
		trace.Debug("time calcurated in VBR");
		
		duration = mad_timer_abs(duration);
		totalTimeSeconds_ = duration.seconds;
		snprintf(timeStr_, 18, "%02u:%02u", static_cast<unsigned int>(duration.seconds / 60), static_cast<unsigned int>(duration.seconds % 60));
	    
		break;
		
	    case XING_VBR:
		trace.Debug("time calculated in Xing header VBR");
		
		mad_timer_multiply(&header.duration, xingFrames_);
		duration = header.duration;
		duration = mad_timer_abs(duration);
		totalTimeSeconds_ = duration.seconds;
		snprintf(timeStr_, 18, "%02u:%02u", static_cast<unsigned int>(duration.seconds / 60), static_cast<unsigned int>(duration.seconds % 60));
	    
		break;		
	}

	totalTime_ = timeStr_;

	mad_frame_finish(&frame);
	mad_header_finish(&header);
	mad_stream_finish(&stream);

	munmap(fdm, stat.st_size);
	
    } else
	trace.Debug("Can't map file");

    close(fd);

}
